from  inspect import getsource
import re
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
#######################################################################################################################
from .rm import *                       # Модификации линейной регрессионной модели
from .cm import *                       # Модификации модели классификации
from .additional_funcs import *         # Дополнительные функции
from .trees import *                    # Деревья


files_dict ={
    'Модификации линейной регрессионной модели': RM,
    'Модификации модели классификации': CM,
    'Дополнительные функции' : AF,
    'Реализация дерева решений' : TREES,
    
}

names = list(files_dict.keys())
modules = list(files_dict.values())

def imports():
    return '''
    
    from scipy.integrate import quad
    import math
    import numpy a np
    import sympy
    import itertools
    sympy.init_printing(use_unicode=True,use_latex=True)
    '''
    
def enable_ppc():
    return'''
import pyperclip

#Делаем функцию которая принимает переменную text
def write(name):
    pyperclip.copy(name) #Копирует в буфер обмена информацию
    pyperclip.paste()'''
    
def invert_dict(d):
    return {value: key for key, value in d.items()}

def get_task_from_func(func,to_search=False):
    return re.search(r'"""\s*(.*?)\s*(?=def __init__|Args)',getsource(func),re.DOTALL).group(0)[3:-4].replace('\n','').replace(' ','') if to_search else re.search(r'"""\s*(.*?)\s*(?=def __init__|Args)',getsource(func),re.DOTALL).group(0)[3:-4]


funcs_dicts = [dict([(get_task_from_func(i), i) for i in module]) for module in modules]
funcs_dicts_ts = [dict([(get_task_from_func(i,True), i) for i in module]) for module in modules]
funcs_dicts_full = [dict([(i.__name__, getsource(i)) for i in module]) for module in modules]


themes_list_funcs = dict([(names[i],list(funcs_dicts[i].values()) ) for i in range(len(names))]) # Название темы : список функций по теме
themes_list_dicts = dict([(names[i],funcs_dicts[i]) for i in range(len(names))])                 # Название темы : словарь по теме, где ЗАДАНИЕ: ФУНКЦИИ
themes_list_dicts_full = dict([(names[i],funcs_dicts_full[i]) for i in range(len(names))])       # Название темы : словарь по теме, где НАЗВАНИЕ ФУНКЦИИ: ТЕКСТ ФУНКЦИИ


# Тема -> Функция -> Задание
def description(dict_to_show = themes_list_funcs, key=None, show_only_keys:bool = False, show_keys_second_level:bool = True, n_symbols:int = 32):
    """
    Печатает информацию о заданиях и функциях 
    
    Parameters
    ----------
    dict_to_show : dict, optional
        словарь, который будет использоваться для поиска заданий, 
        по умолчанию themes_list_funcs
    key : str, optional
        если dict_to_show - строка, то key - это ключ, 
        по которому будет найден словарь в themes_list_dicts_full, 
        если key=None, то будет найден словарь по строке dict_to_show
    show_only_keys : bool, optional
        если True, то будет печататься только список keys, 
        если False, то будет печататься словарь с функциями, 
        по умолчанию False
    show_keys_second_level : bool, optional
        если True, то будет печататься информация о функциях, 
        если False, то будет печататься только список функций, 
        по умолчанию False
    n_symbols : int, optional
        количество символов, которое будет выведено, если show_keys_second_level=True, 
        по умолчанию 20
    
    Returns
    -------
    None
    """
    if dict_to_show=='Вывести функцию буфера обмена':
            return print(enable_ppc)
    
    else:
        if type(dict_to_show) == str and key==None:
                dict_to_show = themes_list_dicts[dict_to_show] # Теперь это словарь ЗАДАНИЕ : ФУНКЦИЯ
                dict_to_show = invert_dict(dict_to_show)       # Теперь это словарь ФУНКЦИЯ : ЗАДАНИЕ
                text = ""
                length1=1+max([len(x.__name__) for x in list(dict_to_show.keys())])
                
                for key in dict_to_show.keys():
                    text += f'{key.__name__:<{length1}}'
                    
                    if not show_only_keys:
                        text +=': '
                        text += f'{dict_to_show[key]};\n'+' '*(length1+2)
                    text += '\n'
                    
                return print(text)
        
        elif type(dict_to_show) == str and key in themes_list_dicts_full[dict_to_show].keys():
            return print(themes_list_dicts_full[dict_to_show][key])
        
        else:
            show_only_keys=False
        text = ""
        length1=1+max([len(x) for x in list(dict_to_show.keys())])
        for key in dict_to_show.keys():
            text += f'{key:^{length1}}'
            if not show_only_keys:
                text +=': '
                for f in dict_to_show[key]:
                    text += f'{f.__name__}'
                    if show_keys_second_level:
                        text += ': '
                        func_text_len = len(invert_dict(themes_list_dicts[key])[f])
                        func_text = invert_dict(themes_list_dicts[key])[f]
                        text += func_text.replace('\n','\n'+' '*(length1 + len(f.__name__))) if func_text_len<n_symbols else func_text[:n_symbols].replace('\n','\n'+' '*(length1 + len(f.__name__)))+'...'
                    text += ';\n'+' '*(length1+2)
            text += '\n'
        return print(text)


from .rrstr import *
#######################################################################################################################
# Непрерывные случайные величины
#######################################################################################################################
def CRV_1(a,b,betas,powers,all_power,q_l=0.9):
    """Абсолютно непрерывная случайная величина  X  может принимать значения только в отрезке [a, b].<br>
    На этом отрезке плотность распределения случайной величины  X  имеет вид:<br> 
    f(x) = C * (1 + sum(betas_i * x ^ powers_i) )^all_power,<br>
    где С - положительная константа.<br>
    Найти:
    - константу C 
    - математическое ожидание E(X)
    - стандартное отклонение σX
    - квантиль уровня k распределения X

    Args:
        a (numerical): Начало отрезка
        b (numerical): Конец отрезка
        betas (array-like): Список коэффициентов при каждом x
        powers (array-like): Список степеней при каждом x
        all_power (numerical): Степень всего выражения
        q_l (numerical, optional): Уровень квантиля. Стандартно равен 0.9

    Returns:
        result (tuple): Ответы на задание по очереди, одновременно с выведением через print()
        
        ### 1 f"C = {C}")
        ### 2 f"E = {f_C.mean()}"
        ### 3 f"sigma = {f_C.std()}"
        ### 4 f"q = {f_C.ppf(q_l)}"
    """
    
    import scipy.stats
    import scipy.integrate
    import sympy
    import numpy as np

    
    assert len(betas)==len(powers), 'Степеней должно быть столько же, сколько коэффициентов Бета'

    # F = f(x) без C    
    F = lambda x: (1+ np.sum([betas[i]*x**powers[i] for i in range(len(betas))]))**all_power

    C = 1 / scipy.integrate.quad(F, a, b)[0]
    def f(x):
        return C * F(x)
    class dist_f_C(scipy.stats.rv_continuous):
    #функция вероятности
        def _pdf(self,x):
            return f(x) if (a<= x <=b) else 0
        #функция значений
        def _expect(self,x):
            return x
    #зададим распределение
    f_C = dist_f_C(a = a, b = b)
    
    #1
    print(f"C = {rrstr(C,5)}")
    #2
    print(f"E = {rrstr(f_C.mean(),3)}")
    #3
    print(f"sigma = {rrstr(f_C.std(),3)}")
    #4
    print(f"q = {rrstr(f_C.ppf(q_l),3)}")
    
    return (C,f_C.mean(),f_C.std(),f_C.ppf(q_l))
#######################################################################################################################
def CRV_2(a,b,betas,powers,all_power,q_l=0.9):
    """Случайная величина X равномерно распределена на отрезке [a,b].<br>
    Случайная величина Y выражается через X следующим образом:<br>
    Y = C * (1 + sum(betas_i * x ^ powers_i) )^all_power<br>
    Найдите:
    - математическое ожидание  E(Y)
    - стандартное отклонение σY
    - асимметрию As(Y)
    - квантиль уровня q_l распределения Y

    Args:
        a (numerical): Начало отрезка
        b (numerical): Конец отрезка
        betas (array-like): Список коэффициентов при каждом x
        powers (array-like): Список степеней при каждом x
        all_power (numerical): Степень всего выражения
        q_l (numerical, optional): Уровень квантиля. Стандартно равен 0.9

    Returns:
        result (tuple): Ответы на задание по очереди, одновременно с выведением через print()
            ### 1 f"EY = {EY}"
            ### 2 "Qy = {Qy}"
            ### 3 f"AsY = {AsY}"
            ### 4 f"q = {q}"
    """
    
    import scipy.stats
    import scipy.integrate
    import sympy
    import numpy as np
    import math
    
    assert len(betas)==len(powers), 'Степеней должно быть столько же, сколько коэффициентов Бета'
    
    F = lambda x: (1+ np.sum([betas[i]*x**powers[i] for i in range(len(betas))]))**all_power
    
    ab = [a, b]
    p = 1/(ab[1]-ab[0])
    EY = p * scipy.integrate.quad(F, a, b)[0]
    
    f = lambda x: (F(x))**2
    VarY = p * scipy.integrate.quad(f, a, b)[0] - EY**2
    Qy = math.sqrt(VarY)
    
    f = lambda x: p * (F(x) - EY)**3
    AsY = scipy.integrate.quad(f, a, b)[0]/Qy**3
    
    q = F(ab[0] + (ab[1] - ab[0])*q_l)
    
    #1
    print(f"EY = {rrstr(EY,1)}")
    #2
    print(f"Qy = {rrstr(Qy,2)}")
    #3
    print(f"AsY = {rrstr(AsY,4)}")
    #4
    print(f"q = {rrstr(q,3)}")
    
    return (EY,Qy,AsY,q)
#######################################################################################################################
CRV = [CRV_1,CRV_2]